import typing


class DatabaseConfig:
    """The configuration to be used for the :ref:`royalnet.database.Alchemy` component of :ref:`royalnet.bots.GenericBot`."""

    def __init__(self,
                 database_uri: str,
                 master_table: typing.Type,
                 identity_table: typing.Type,
                 identity_column_name: str):
        self.database_uri: str = database_uri
        self.master_table: typing.Type = master_table
        self.identity_table: typing.Type = identity_table
        self.identity_column_name: str = identity_column_name



import asyncio
import websockets
import uuid
import functools
import typing
import pickle
import logging as _logging
from .messages import Message, ServerErrorMessage, RequestError
from .packages import Package

default_loop = asyncio.get_event_loop()
log = _logging.getLogger(__name__)


class NotConnectedError(Exception):
    """The :py:class:`royalnet.network.RoyalnetLink` is not connected to a :py:class:`royalnet.network.RoyalnetServer`."""


class NotIdentifiedError(Exception):
    """The :py:class:`royalnet.network.RoyalnetLink` has not identified yet to a :py:class:`royalnet.network.RoyalnetServer`."""


class NetworkError(Exception):
    def __init__(self, error_msg: ServerErrorMessage, *args):
        super().__init__(*args)
        self.error_msg: ServerErrorMessage = error_msg


class PendingRequest:
    def __init__(self, *, loop=default_loop):
        self.event: asyncio.Event = asyncio.Event(loop=loop)
        self.data: typing.Optional[Message] = None

    def __repr__(self):
        if self.event.is_set():
            return f"<PendingRequest: {self.data.__class__.__name__}>"
        return f"<PendingRequest>"

    def set(self, data):
        self.data = data
        self.event.set()


def requires_connection(func):
    @functools.wraps(func)
    async def new_func(self, *args, **kwargs):
        await self._connect_event.wait()
        return await func(self, *args, **kwargs)
    return new_func


def requires_identification(func):
    @functools.wraps(func)
    async def new_func(self, *args, **kwargs):
        await self.identify_event.wait()
        return await func(self, *args, **kwargs)
    return new_func


class RoyalnetLink:
    def __init__(self, master_uri: str, secret: str, link_type: str, request_handler, *,
                 loop: asyncio.AbstractEventLoop = default_loop):
        assert ":" not in link_type
        self.master_uri: str = master_uri
        self.link_type: str = link_type
        self.nid: str = str(uuid.uuid4())
        self.secret: str = secret
        self.websocket: typing.Optional[websockets.WebSocketClientProtocol] = None
        self.request_handler = request_handler
        self._pending_requests: typing.Dict[str, typing.Optional[Message]] = {}
        self._loop: asyncio.AbstractEventLoop = loop
        self._connect_event: asyncio.Event = asyncio.Event(loop=self._loop)
        self.identify_event: asyncio.Event = asyncio.Event(loop=self._loop)

    async def connect(self):
        log.info(f"Connecting to {self.master_uri}...")
        self.websocket = await websockets.connect(self.master_uri, loop=self._loop)
        self._connect_event.set()
        log.info(f"Connected!")

    @requires_connection
    async def receive(self) -> Package:
        try:
            raw_pickle = await self.websocket.recv()
        except websockets.ConnectionClosed:
            self.websocket = None
            self._connect_event.clear()
            self.identify_event.clear()
            log.info(f"Connection to {self.master_uri} was closed.")
            # What to do now? Let's just reraise.
            raise
        package: typing.Union[Package, Package] = pickle.loads(raw_pickle)
        assert package.destination == self.nid
        log.debug(f"Received package: {package}")
        return package

    @requires_connection
    async def identify(self) -> None:
        log.info(f"Identifying to {self.master_uri}...")
        await self.websocket.send(f"Identify {self.nid}:{self.link_type}:{self.secret}")
        response_package = await self.receive()
        response = response_package.data
        if isinstance(response, ServerErrorMessage):
            raise NetworkError(response, "Server returned error while identifying self")
        self.identify_event.set()
        log.info(f"Identified successfully!")

    @requires_identification
    async def send(self, package: Package):
        raw_pickle: bytes = pickle.dumps(package)
        await self.websocket.send(raw_pickle)
        log.debug(f"Sent package: {package}")

    @requires_identification
    async def request(self, message, destination):
        package = Package(message, destination, self.nid)
        request = PendingRequest(loop=self._loop)
        self._pending_requests[package.source_conv_id] = request
        await self.send(package)
        log.debug(f"Sent request: {message} -> {destination}")
        await request.event.wait()
        result: Message = request.data
        log.debug(f"Received response: {request} -> {result}")
        if isinstance(result, ServerErrorMessage):
            raise NetworkError(result, "Server returned error while requesting something")
        return result

    async def run(self):
        log.debug(f"Running main client loop for {self.nid}.")
        while True:
            if self.websocket is None:
                await self.connect()
            if not self.identify_event.is_set():
                await self.identify()
            package: Package = await self.receive()
            # Package is a response
            if package.destination_conv_id in self._pending_requests:
                request = self._pending_requests[package.destination_conv_id]
                request.set(package.data)
                continue
            # Package is a request
            assert isinstance(package, Package)
            log.debug(f"Received request {package.source_conv_id}: {package}")
            try:
                response = await self.request_handler(package.data)
                assert isinstance(response, Message)
            except Exception as exc:
                response = RequestError(exc=exc)
            response_package: Package = package.reply(response)
            await self.send(response_package)
            log.debug(f"Replied to request {response_package.source_conv_id}: {response_package}")



import re
import typing
from royalnet.error import InvalidInputError


class CommandArgs(list):
    """The arguments of a command."""

    def __getitem__(self, item):
        """Arguments can be accessed with an array notation, such as ``args[0]``.

        Raises:
             :py:exc:`InvalidInputError` if the requested argument does not exist."""
        if isinstance(item, int):
            try:
                return super().__getitem__(item)
            except IndexError:
                raise InvalidInputError(f'Tried to get missing [{item}] arg from CommandArgs')
        if isinstance(item, slice):
            try:
                return super().__getitem__(item)
            except IndexError:
                raise InvalidInputError(f'Tried to get invalid [{item}] slice from CommandArgs')
        raise ValueError(f"Invalid type passed to CommandArgs.__getattr__: {type(item)}")

    def joined(self, *, require_at_least=0) -> str:
        """Get the arguments as a space-joined string.

        Parameters:
            require_at_least: the minimum amount of arguments required, will raise :py:exc:`royalnet.error.InvalidInputError` if the requirement is not fullfilled.

        Raises:
            :py:exc:`royalnet.error.InvalidInputError` if the ``require_at_least`` requirement is not fullfilled.

        Returns:
            The space-joined string."""
        if len(self) < require_at_least:
            raise InvalidInputError("Not enough arguments")
        return " ".join(self)

    def match(self, pattern: typing.Pattern) -> typing.Sequence[typing.AnyStr]:
        """Match the :py:func:`royalnet.utils.commandargs.joined` to a regex pattern.

        Parameters:
            pattern: The regex pattern to be passed to :py:func:`re.match`.

        Returns:
            The matched groups, as returned by :py:func:`re.Match.groups`."""
        text = self.joined()
        match = re.match(pattern, text)
        if match is None:
            raise InvalidInputError("Pattern didn't match")
        return match.groups()

    def optional(self, index: int, default=None):
        """Get the argument at a specific index, but don't raise an error if nothing is found, instead returning the ``default`` value.

        Parameters:
            index: The index of the argument you want to retrieve.
            default: The value returned if the argument is missing.

        Returns:
            Either the argument or the ``default`` value, defaulting to ``None``."""
        try:
            return self[index]
        except InvalidInputError:
            return default



import typing
from ..error import UnsupportedError
if typing.TYPE_CHECKING:
    from .call import Call
    from ..utils import NetworkHandler


class Command:
    """A generic command, called from any source."""

    command_name: str = NotImplemented
    command_description: str = NotImplemented
    command_syntax: str = NotImplemented

    require_alchemy_tables: typing.Set = set()

    network_handlers: typing.List[typing.Type["NetworkHandler"]] = {}

    @classmethod
    async def common(cls, call: "Call"):
        raise UnsupportedError()

    @classmethod
    def network_handler_dict(cls):
        d = {}
        for network_handler in cls.network_handlers:
            d[network_handler.message_type] = network_handler
        return d



import typing
import asyncio
from ..network import Message, Reply
from .command import Command
from .commandargs import CommandArgs
if typing.TYPE_CHECKING:
    from ..database import Alchemy


loop = asyncio.get_event_loop()


class Call:
    """A command call. An abstract class, sub-bots should create a new call class from this."""

    # These parameters / methods should be overridden
    interface_name = NotImplemented
    interface_obj = NotImplemented
    interface_prefix = NotImplemented
    alchemy: "Alchemy" = NotImplemented

    async def reply(self, text: str) -> None:
        """Send a text message to the channel where the call was made.

        Parameters:
             text: The text to be sent, possibly formatted in the weird undescribed markup that I'm using."""
        raise NotImplementedError()

    async def net_request(self, message, destination: str) -> Reply:
        """Send data through a :py:class:`royalnet.network.RoyalnetLink` and wait for a :py:class:`royalnet.network.Reply`.

        Parameters:
            message: The data to be sent. Must be :py:mod:`pickle`-able.
            destination: """
        raise NotImplementedError()

    async def get_author(self, error_if_none=False):
        """Try to find the universal identifier of the user that sent the message.
        That probably means, the database row identifying the user.

        Parameters:
            error_if_none: Raise a :py:exc:`royalnet.error.UnregisteredError` if this is True and the call has no author.

        Raises:
             :py:exc:`royalnet.error.UnregisteredError` if ``error_if_none`` is set to True and no author is found."""
        raise NotImplementedError()

    # These parameters / methods should be left alone
    def __init__(self,
                 channel,
                 command: typing.Type[Command],
                 command_args: typing.List[str] = None,
                 **kwargs):
        """Create the call.

        Parameters:
            channel: The channel object this call was sent in.
            command: The command to be called.
            command_args: The arguments to be passed to the command
            kwargs: Additional optional keyword arguments that may be passed to the command, possibly specific to the bot.
            """
        if command_args is None:
            command_args = []
        self.channel = channel
        self.command = command
        self.args = CommandArgs(command_args)
        self.kwargs = kwargs
        self.session = None

    async def _session_init(self):
        """If the command requires database access, create a :py:class:`royalnet.database.Alchemy` session for this call, otherwise, do nothing."""
        if not self.command.require_alchemy_tables:
            return
        self.session = await loop.run_in_executor(None, self.alchemy.Session)

    async def session_end(self):
        """Close the previously created :py:class:`royalnet.database.Alchemy` session for this call (if it was created)."""
        if not self.session:
            return
        self.session.close()

    async def run(self):
        """Execute the called command, and return the command result."""
        await self._session_init()
        try:
            coroutine = getattr(self.command, self.interface_name)
        except AttributeError:
            coroutine = self.command.common
        try:
            result = await coroutine(self)
        finally:
            await self.session_end()
        return result



import pickle
import uuid


class Package:
    """A Royalnet package, the data type with which a :py:class:`royalnet.network.RoyalnetLink` communicates with a :py:class:`royalnet.network.RoyalnetServer` or another link. """

    def __init__(self, data, destination: str, source: str, *, source_conv_id: str = None, destination_conv_id: str = None):
        """Create a Package.

        Parameters:
            data: The data that should be sent. Usually a :py:class:`royalnet.network.Message`.
            destination: The ``link_type`` of the destination node, or alternatively, the ``nid`` of the node. Can also be the ``NULL`` value to send the message to nobody.
            source: The ``nid`` of the node that created this Package.
            source_conv_id: The conversation id of the node that created this package. Akin to the sequence number on IP packets.
            destination_conv_id: The conversation id of the node that this Package is a reply to."""
        # TODO: something is not right in these type hints. Check them.
        self.data = data
        self.destination: str = destination
        self.source: str = source
        self.source_conv_id: str = source_conv_id or str(uuid.uuid4())
        self.destination_conv_id: str = destination_conv_id

    def __repr__(self):
        return f"<Package to {self.destination}: {self.data.__class__.__name__}>"

    def reply(self, data) -> "Package":
        """Reply to this Package with another Package.

        Parameters:
            data: The data that should be sent. Usually a :py:class:`royalnet.network.Message`.

        Returns:
            The reply Package."""
        return Package(data, self.source, self.destination,
                       source_conv_id=str(uuid.uuid4()),
                       destination_conv_id=self.source_conv_id)

    def pickle(self) -> bytes:
        """:py:mod:`pickle` this Package.

        Returns:
            The pickled package in form of bytes."""
        return pickle.dumps(self)



import asyncio
import logging as _logging
from ..utils import Command, Call

loop = asyncio.get_event_loop()
log = _logging.getLogger(__name__)


class MissingCommand(Command):

    command_name = "missing"
    command_description = "Informa che il comando non esiste."
    command_syntax = ""

    @classmethod
    async def common(cls, call: Call):
        await call.reply(f"⚠️ Il comando richiesto non esiste.")



import discord
import asyncio
import typing
import logging as _logging
from .generic import GenericBot
from ..commands import NullCommand
from ..utils import asyncify, Call, Command
from ..error import UnregisteredError, NoneFoundError, TooManyFoundError, InvalidConfigError
from ..network import Message, Reply, RoyalnetConfig
from ..database import DatabaseConfig
from ..audio import PlayMode, Playlist, RoyalPCMAudio

loop = asyncio.get_event_loop()
log = _logging.getLogger(__name__)

# TODO: Load the opus library
if not discord.opus.is_loaded():
    log.error("Opus is not loaded. Weird behaviour might emerge.")


class DiscordConfig:
    """The specific configuration to be used for :py:class:`royalnet.bots.DiscordBot`."""
    def __init__(self, token: str):
        self.token = token


class DiscordBot(GenericBot):
    """A bot that connects to `Discord <https://discordapp.com/>`_."""

    interface_name = "discord"

    def _init_voice(self):
        """Initialize the variables needed for the connection to voice chat."""
        log.debug(f"Creating music_data dict")
        self.music_data: typing.Dict[discord.Guild, PlayMode] = {}

    def _call_factory(self) -> typing.Type[Call]:
        log.debug(f"Creating DiscordCall")

        # noinspection PyMethodParameters
        class DiscordCall(Call):
            interface_name = self.interface_name
            interface_obj = self
            interface_prefix = "!"

            alchemy = self.alchemy

            async def reply(call, text: str):
                # TODO: don't escape characters inside [c][/c] blocks
                escaped_text = text.replace("*", "\\*") \
                    .replace("_", "\\_") \
                    .replace("`", "\\`") \
                    .replace("[b]", "**") \
                    .replace("[/b]", "**") \
                    .replace("[i]", "_") \
                    .replace("[/i]", "_") \
                    .replace("[u]", "__") \
                    .replace("[/u]", "__") \
                    .replace("[c]", "`") \
                    .replace("[/c]", "`") \
                    .replace("[p]", "```") \
                    .replace("[/p]", "```")
                await call.channel.send(escaped_text)

            async def net_request(call, message: Message, destination: str):
                if self.network is None:
                    raise InvalidConfigError("Royalnet is not enabled on this bot")
                response: Reply = await self.network.request(message, destination)
                response.raise_on_error()
                return response

            async def get_author(call, error_if_none=False):
                message: discord.Message = call.kwargs["message"]
                user: discord.Member = message.author
                query = call.session.query(self.master_table)
                for link in self.identity_chain:
                    query = query.join(link.mapper.class_)
                query = query.filter(self.identity_column == user.id)
                result = await asyncify(query.one_or_none)
                if result is None and error_if_none:
                    raise UnregisteredError("Author is not registered")
                return result

        return DiscordCall

    def _bot_factory(self) -> typing.Type[discord.Client]:
        """Create a custom DiscordClient class inheriting from :py:class:`discord.Client`."""
        log.debug(f"Creating DiscordClient")

        # noinspection PyMethodParameters
        class DiscordClient(discord.Client):
            async def vc_connect_or_move(cli, channel: discord.VoiceChannel):
                # Connect to voice chat
                try:
                    await channel.connect()
                    log.debug(f"Connecting to Voice in {channel}")
                except discord.errors.ClientException:
                    # Move to the selected channel, instead of connecting
                    # noinspection PyUnusedLocal
                    for voice_client in cli.voice_clients:
                        voice_client: discord.VoiceClient
                        if voice_client.guild != channel.guild:
                            continue
                        await voice_client.move_to(channel)
                        log.debug(f"Moved {voice_client} to {channel}")
                # Create a music_data entry, if it doesn't exist; default is a Playlist
                if not self.music_data.get(channel.guild):
                    log.debug(f"Creating music_data for {channel.guild}")
                    self.music_data[channel.guild] = Playlist()

            @staticmethod
            async def on_message(message: discord.Message):
                text = message.content
                # Skip non-text messages
                if not text:
                    return
                # Skip bot messages
                author: typing.Union[discord.User] = message.author
                if author.bot:
                    return
                # Find and clean parameters
                command_text, *parameters = text.split(" ")
                # Don't use a case-sensitive command name
                command_name = command_text.lower()
                # Call the command
                await self.call(command_name, message.channel, parameters, message=message)

            async def on_ready(cli):
                log.debug("Connection successful, client is ready")
                await cli.change_presence(status=discord.Status.online)

            def find_guild_by_name(cli, name: str) -> discord.Guild:
                """Find the :py:class:`discord.Guild` with the specified name. Case-insensitive.

                Raises:
                     :py:exc:`NoneFoundError` if no channels are found.
                     :py:exc:`TooManyFoundError` if more than one is found."""
                all_guilds: typing.List[discord.Guild] = cli.guilds
                matching_channels: typing.List[discord.Guild] = []
                for guild in all_guilds:
                    if guild.name.lower() == name.lower():
                        matching_channels.append(guild)
                if len(matching_channels) == 0:
                    raise NoneFoundError("No channels were found")
                elif len(matching_channels) > 1:
                    raise TooManyFoundError("Too many channels were found")
                return matching_channels[0]

            def find_channel_by_name(cli,
                                     name: str,
                                     guild: typing.Optional[discord.Guild] = None) -> discord.abc.GuildChannel:
                """Find the :py:class:`TextChannel`, :py:class:`VoiceChannel` or :py:class:`CategoryChannel` with the specified name. Case-insensitive.
                Guild is optional, but the method will raise a :py:exc:`TooManyFoundError` if none is specified and there is more than one channel with the same name.
                Will also raise a :py:exc:`NoneFoundError` if no channels are found."""
                if guild is not None:
                    all_channels = guild.channels
                else:
                    all_channels: typing.List[discord.abc.GuildChannel] = cli.get_all_channels()
                matching_channels: typing.List[discord.abc.GuildChannel] = []
                for channel in all_channels:
                    if not (isinstance(channel, discord.TextChannel)
                            or isinstance(channel, discord.VoiceChannel)
                            or isinstance(channel, discord.CategoryChannel)):
                        continue
                    if channel.name.lower() == name.lower():
                        matching_channels.append(channel)
                if len(matching_channels) == 0:
                    raise NoneFoundError("No channels were found")
                elif len(matching_channels) > 1:
                    raise TooManyFoundError("Too many channels were found")
                return matching_channels[0]

            def find_voice_client_by_guild(cli, guild: discord.Guild):
                """Find the :py:class:`discord.VoiceClient` belonging to a specific :py:class:`discord.Guild`.

                Raises:
                     :py:exc:`NoneFoundError` if the :py:class:`discord.Guild` currently has no :py:class:`discord.VoiceClient`."""
                for voice_client in cli.voice_clients:
                    if voice_client.guild == guild:
                        return voice_client
                raise NoneFoundError("No voice clients found")

        return DiscordClient

    def _init_client(self):
        """Create an instance of the DiscordClient class created in :py:func:`royalnet.bots.DiscordBot._bot_factory`."""
        log.debug(f"Creating DiscordClient instance")
        self._Client = self._bot_factory()
        self.client = self._Client()

    def __init__(self, *,
                 discord_config: DiscordConfig,
                 royalnet_config: typing.Optional[RoyalnetConfig] = None,
                 database_config: typing.Optional[DatabaseConfig] = None,
                 command_prefix: str = "!",
                 commands: typing.List[typing.Type[Command]] = None,
                 missing_command: typing.Type[Command] = NullCommand,
                 error_command: typing.Type[Command] = NullCommand):
        super().__init__(royalnet_config=royalnet_config,
                         database_config=database_config,
                         command_prefix=command_prefix,
                         commands=commands,
                         missing_command=missing_command,
                         error_command=error_command)
        self._discord_config = discord_config
        self._init_client()
        self._init_voice()

    async def run(self):
        """Login to Discord, then run the bot."""
        log.debug(f"Logging in to Discord")
        await self.client.login(self._discord_config.token)
        log.debug(f"Connecting to Discord")
        await self.client.connect()
        # TODO: how to stop?

    async def add_to_music_data(self, audio_sources: typing.List[RoyalPCMAudio], guild: discord.Guild):
        """Add a file to the corresponding music_data object."""
        guild_music_data = self.music_data[guild]
        for audio_source in audio_sources:
            log.debug(f"Adding {audio_source} to music_data")
            guild_music_data.add(audio_source)
        if guild_music_data.now_playing is None:
            await self.advance_music_data(guild)

    async def advance_music_data(self, guild: discord.Guild):
        """Try to play the next song, while it exists. Otherwise, just return."""
        guild_music_data = self.music_data[guild]
        voice_client = self.client.find_voice_client_by_guild(guild)
        next_source: RoyalPCMAudio = await guild_music_data.next()
        await self.update_activity_with_source_title()
        if next_source is None:
            log.debug(f"Ending playback chain")
            return

        def advance(error=None):
            if error:
                raise Exception(f"Error while advancing music_data: {error}")
            loop.create_task(self.advance_music_data(guild))

        log.debug(f"Starting playback of {next_source}")
        voice_client.play(next_source, after=advance)

    async def update_activity_with_source_title(self):
        """Change the bot's presence (using :py:func:`discord.Client.change_presence`) to match the current listening status.

        If multiple guilds are using the bot, the bot will always have an empty presence."""
        if len(self.music_data) != 1:
            # Multiple guilds are using the bot, do not display anything
            log.debug(f"Updating current Activity: setting to None, as multiple guilds are using the bot")
            await self.client.change_presence(status=discord.Status.online)
            return
        # FIXME: PyCharm faulty inspection?
        # noinspection PyUnresolvedReferences
        play_mode: PlayMode = list(self.music_data.items())[0][1]
        now_playing = play_mode.now_playing
        if now_playing is None:
            # No songs are playing now
            log.debug(f"Updating current Activity: setting to None, as nothing is currently being played")
            await self.client.change_presence(status=discord.Status.online)
            return
        log.debug(f"Updating current Activity: listening to {now_playing.rpf.info.title}")
        await self.client.change_presence(activity=discord.Activity(name=now_playing.rpf.info.title,
                                                                    type=discord.ActivityType.listening),
                                          status=discord.Status.online)



"""Relational database classes and methods."""

from .alchemy import Alchemy
from .relationshiplinkchain import relationshiplinkchain
from .databaseconfig import DatabaseConfig

__all__ = ["Alchemy", "relationshiplinkchain", "DatabaseConfig"]



import typing
import asyncio
import youtube_dl
import ffmpeg
from ..utils import Command, Call, NetworkHandler, asyncify
from ..network import Message, RequestSuccessful
from ..error import TooManyFoundError, NoneFoundError
from ..audio import RoyalPCMAudio, YtdlInfo
if typing.TYPE_CHECKING:
    from ..bots import DiscordBot


loop = asyncio.get_event_loop()


class PlayMessage(Message):
    def __init__(self, url: str, guild_name: typing.Optional[str] = None):
        self.url: str = url
        self.guild_name: typing.Optional[str] = guild_name


class PlaySuccessful(RequestSuccessful):
    def __init__(self, info_list: typing.List[YtdlInfo]):
        self.info_list: typing.List[YtdlInfo] = info_list


class PlayNH(NetworkHandler):
    message_type = PlayMessage

    @classmethod
    async def discord(cls, bot: "DiscordBot", message: PlayMessage):
        """Handle a play Royalnet request. That is, add audio to a PlayMode."""
        # Find the matching guild
        if message.guild_name:
            guild = bot.client.find_guild(message.guild_name)
        else:
            if len(bot.music_data) == 0:
                raise NoneFoundError("No voice clients active")
            if len(bot.music_data) > 1:
                raise TooManyFoundError("Multiple guilds found")
            guild = list(bot.music_data)[0]
        # Ensure the guild has a PlayMode before adding the file to it
        if not bot.music_data.get(guild):
            # TODO: change Exception
            raise Exception("No music_data for this guild")
        # Start downloading
        if message.url.startswith("http://") or message.url.startswith("https://"):
            audio_sources: typing.List[RoyalPCMAudio] = await asyncify(RoyalPCMAudio.create_from_url, message.url)
        else:
            audio_sources = await asyncify(RoyalPCMAudio.create_from_ytsearch, message.url)
        await bot.add_to_music_data(audio_sources, guild)
        return PlaySuccessful(info_list=[source.rpf.info for source in audio_sources])


async def notify_on_timeout(call: Call, url: str, time: float, repeat: bool = False):
    """Send a message after a while to let the user know that the bot is still downloading the files and hasn't crashed."""
    while True:
        await asyncio.sleep(time)
        await call.reply(f"ℹ️ Il download di [c]{url}[/c] sta richiedendo più tempo del solito, ma è ancora in corso!")
        if not repeat:
            break


class PlayCommand(Command):
    command_name = "play"
    command_description = "Riproduce una canzone in chat vocale."
    command_syntax = "[ [guild] ] (url)"

    network_handlers = [PlayNH]

    @classmethod
    async def common(cls, call: Call):
        guild, url = call.args.match(r"(?:\[(.+)])?\s*(.+)")
        download_task = loop.create_task(call.net_request(PlayMessage(url, guild), "discord"))
        notify_task = loop.create_task(notify_on_timeout(call, url, time=30, repeat=True))
        try:
            response: PlaySuccessful = await download_task
        except Exception as exc:
            # RoyalPCMFile errors
            if isinstance(exc, FileExistsError):
                await call.reply(f"❌ Scaricare [c]{url}[/c] significherebbe sovrascrivere un file già esistente.\nQuesto è un bug, e non dovrebbe mai succedere. Se è appena successo, segnalate il problema a https://github.com/Steffo99/royalnet/issues.\n[p]{exc}[/p]")
                return
            # ffmpeg errors
            if isinstance(exc, ffmpeg.Error):
                await call.reply(f"⚠️ Errore durante la conversione a PCM di [c]{url}[/c]:\n[p]{exc}[/p]")
                return
            # youtube_dl errors
            if isinstance(exc, youtube_dl.utils.ContentTooShortError):
                await call.reply(f"⚠️ Mentre era in corso il download di [c]{url}[/c], la connessione è stata interrotta, quindi la riproduzione è stata annullata.\n[p]{exc}[/p]")
                return
            if isinstance(exc, youtube_dl.utils.UnavailableVideoError):
                await call.reply(f"⚠️ Non è disponibile nessun audio su [c]{url}[/c].\n[p]{exc}[/p]")
                return
            if isinstance(exc, youtube_dl.utils.SameFileError):
                await call.reply(f"❌ Scaricare [c]{url}[/c] significherebbe scaricare due file diversi sullo stesso nome.\nQuesto è un bug, e non dovrebbe mai succedere. Se è appena successo, segnalate il problema a https://github.com/Steffo99/royalnet/issues.\n[p]{exc}[/p]")
                return
            if isinstance(exc, youtube_dl.utils.GeoRestrictedError):
                await call.reply(f"⚠️ [c]{url}[/c] non può essere visualizzato nel paese in cui si trova il bot e non può essere scaricato.\n[p]{exc}[/p]")
                return
            if isinstance(exc, youtube_dl.utils.UnsupportedError):
                await call.reply(f"⚠️ [c]{url}[/c] non è supportato da YoutubeDL e non può essere scaricato.\n[p]{exc}[/p]")
                return
            if isinstance(exc, youtube_dl.utils.ExtractorError):
                await call.reply(f"⚠️ Errore nella ricerca di info per [c]{url}[/c]:\n[p]{exc}[/p]")
                return
            if isinstance(exc, youtube_dl.utils.DownloadError):
                await call.reply(f"⚠️ Errore nel download di [c]{url}[/c]:\n[p]{exc}[/p]")
                return
            if isinstance(exc, youtube_dl.utils.YoutubeDLError):
                await call.reply(f"⚠️ Errore di youtube_dl per [c]{url}[/c]:\n[p]{exc}[/p]")
                return
            raise
        finally:
            notify_task.cancel()
        for info in response.info_list:
            await call.reply(f"⬇️ Download di [i]{info.title}[/i] completato.")



from ..utils import Command, Call
from telegram import Update, User


class CiaoruoziCommand(Command):

    command_name = "ciaoruozi"
    command_description = "Saluta Ruozi, anche se non è più in RYG."
    command_syntax = ""

    @classmethod
    async def common(cls, call: "Call"):
        await call.reply("👋 Ciao Ruozi!")

    @classmethod
    async def telegram(cls, call: Call):
        update: Update = call.kwargs["update"]
        user: User = update.effective_user
        if user.id == 112437036:
            await call.reply("👋 Ciao me!")
        else:
            await call.reply("👋 Ciao Ruozi!")



import logging
import ffmpeg
import os
import typing
import time
from .youtubedl import YtdlFile, YtdlInfo
from ..utils import safefilename


log = logging.getLogger(__name__)


class RoyalPCMFile(YtdlFile):
    ytdl_args = {
        "format": "bestaudio"  # Fetch the best audio format available
    }

    def __init__(self, info: "YtdlInfo", **ytdl_args):
        # Preemptively initialize info to be able to generate the filename
        self.info = info
        # Set the time to generate the filename
        self._time = time.time()
        # Ensure the file doesn't already exist
        if os.path.exists(self.ytdl_filename) or os.path.exists(self.audio_filename):
            raise FileExistsError("Can't overwrite file")
        # Overwrite the new ytdl_args
        self.ytdl_args = {**self.ytdl_args, **ytdl_args}
        log.info(f"Now downloading {info.webpage_url}")
        super().__init__(info, outtmpl=self.ytdl_filename, **self.ytdl_args)
        # Find the audio_filename with a regex (should be video.opus)
        log.info(f"Converting {self.video_filename}...")
        # Convert the video to pcm
        try:
            ffmpeg.input(f"./{self.video_filename}") \
                  .output(self.audio_filename, format="s16le", ac=2, ar="48000") \
                  .overwrite_output() \
                  .run(quiet=False)
        except ffmpeg.Error as exc:
            log.error(f"FFmpeg error: {exc.stderr}")
            raise
        # Delete the video file
        log.info(f"Deleting {self.video_filename}")
        self.delete_video_file()

    def __repr__(self):
        return f"<RoyalPCMFile {self.audio_filename}>"

    @staticmethod
    def create_from_url(url: str, **ytdl_args) -> typing.List["RoyalPCMFile"]:
        """Download a file with youtube_dl and create a list of :py:class:`discord.audio.RoyalPCMFile`.

        Parameters:
            url: The url of the file to download.
            ytdl_args: Extra arguments to be passed to YoutubeDL while downloading.

        Returns:
            A :py:class:`list` of RoyalPCMAudios, each corresponding to a downloaded video."""
        info_list = YtdlInfo.create_from_url(url)
        return [RoyalPCMFile(info, **ytdl_args) for info in info_list]

    @staticmethod
    def create_from_ytsearch(search: str, amount: int = 1, **ytdl_args) -> typing.List["RoyalPCMFile"]:
        """Search a string on YouTube and download the first ``amount`` number of videos, then download those with youtube_dl and create a list of :py:class:`discord.audio.RoyalPCMFile`.

        Parameters:
            search: The string to search on YouTube.
            amount: The number of videos to download.
            ytdl_args: Extra arguments to be passed to YoutubeDL while downloading.

        Returns:
            A :py:class:`list` of RoyalPCMFiles, each corresponding to a downloaded video."""
        url = f"ytsearch{amount}:{search}"
        info_list = YtdlInfo.create_from_url(url)
        return [RoyalPCMFile(info, **ytdl_args) for info in info_list]

    @property
    def ytdl_filename(self) -> str:
        """
        Returns:
            The name of the downloaded video file, as a :py:class:`str`.

        Warning:
            It's going to be deleted as soon as the :py:func:`royalnet.audio.RoyalPCMFile.__init__` function has completed, so it's probably not going to be very useful...
        """
        return f"./downloads/{safefilename(self.info.title)}-{safefilename(str(int(self._time)))}.ytdl"

    @property
    def audio_filename(self) -> str:
        """
        Returns:
            The name of the downloaded and PCM-converted audio file."""
        return f"./downloads/{safefilename(self.info.title)}-{safefilename(str(int(self._time)))}.pcm"

    def delete_audio_file(self):
        """Delete the PCM-converted audio file."""
        log.info(f"Deleting {self.audio_filename}")
        os.remove(self.audio_filename)



def cdj(class_) -> dict:
    """Return a dict of the class attributes without the ``__module__``, ``__dict__``, ``__weakref__`` and ``__doc__`` keys, to be used while generating dynamically SQLAlchemy declarative table classes."""
    d = dict(class_.__dict__)
    del d["__module__"]
    del d["__dict__"]
    del d["__weakref__"]
    del d["__doc__"]
    return d



import re
from ..utils import Command, Call, safeformat


SHIP_RESULT = "💕 {one} + {two} = [b]{result}[/b]"


class ShipCommand(Command):

    command_name = "ship"
    command_description = "Crea una ship tra due cose."
    command_syntax = "(uno) (due)"

    @classmethod
    async def common(cls, call: Call):
        name_one = call.args[0]
        name_two = call.args[1]
        if name_two == "+":
            name_two = call.args[2]
        name_one = name_one.lower()
        name_two = name_two.lower()
        # Get all letters until the first vowel, included
        match_one = re.search(r"^[A-Za-z][^aeiouAEIOU]*[aeiouAEIOU]?", name_one)
        if match_one is None:
            part_one = name_one[:int(len(name_one) / 2)]
        else:
            part_one = match_one.group(0)
        # Get all letters from the second to last vowel, excluded
        match_two = re.search(r"[^aeiouAEIOU]*[aeiouAEIOU]?[A-Za-z]$", name_two)
        if match_two is None:
            part_two = name_two[int(len(name_two) / 2):]
        else:
            part_two = match_two.group(0)
        # Combine the two name parts
        mixed = part_one + part_two
        await call.reply(safeformat(SHIP_RESULT,
                                    one=name_one.capitalize(),
                                    two=name_two.capitalize(),
                                    result=mixed.capitalize()))



import typing
import asyncio
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from contextlib import contextmanager, asynccontextmanager
from ..utils import asyncify
# noinspection PyUnresolvedReferences
from ..error import InvalidConfigError

loop = asyncio.get_event_loop()


class Alchemy:
    """A wrapper around SQLAlchemy declarative that allows to use multiple databases at once while maintaining a single table-class for both of them."""

    def __init__(self, database_uri: str, tables: typing.Set):
        """Create a new Alchemy object.

        Args:
            database_uri: The uri of the database, as described at https://docs.sqlalchemy.org/en/13/core/engines.html .
            tables: The set of tables to be created and used in the selected database. Check the tables submodule for more details.
        """
        if database_uri.startswith("sqlite"):
            raise NotImplementedError("Support for sqlite databases is currently missing")
        self.engine = create_engine(database_uri)
        self.Base = declarative_base(bind=self.engine)
        self.Session = sessionmaker(bind=self.engine)
        self._create_tables(tables)

    def _create_tables(self, tables: typing.Set):
        for table in tables:
            name = table.__name__
            try:
                self.__getattribute__(name)
            except AttributeError:
                # Actually the intended result
                # TODO: here is the problem!
                self.__setattr__(name, type(name, (self.Base, table), {}))
            else:
                raise NameError(f"{name} is a reserved name and can't be used as a table name")
        self.Base.metadata.create_all()

    @contextmanager
    def session_cm(self):
        """Use Alchemy as a context manager (to be used in with statements)."""
        session = self.Session()
        try:
            yield session
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()

    @asynccontextmanager
    async def session_acm(self):
        """Use Alchemy as a asyncronous context manager (to be used in async with statements)."""
        session = await asyncify(self.Session)
        try:
            yield session
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()



from sqlalchemy import Column, \
                       Integer, \
                       String, \
                       ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declared_attr
# noinspection PyUnresolvedReferences
from .royals import Royal


class Alias:
    __tablename__ = "aliases"

    @declared_attr
    def royal_id(self):
        return Column(Integer, ForeignKey("royals.uid"))

    @declared_attr
    def alias(self):
        return Column(String, primary_key=True)

    @declared_attr
    def royal(self):
        return relationship("Royal", backref="aliases")

    def __repr__(self):
        return f"<Alias {str(self)}>"

    def __str__(self):
        return f"{self.alias}->{self.royal_id}"



class RoyalnetConfig:
    def __init__(self,
                 master_uri: str,
                 master_secret: str):
        if not (master_uri.startswith("ws://")
                or master_uri.startswith("wss://")):
            raise ValueError("Invalid protocol (must be ws:// or wss://)")
        self.master_uri = master_uri
        self.master_secret = master_secret



"""Commands that can be used in bots. These probably won't suit your needs, as they are tailored for the bots of the `Royal Games <https://ryg.steffo.eu>`_ gaming community, but you can check them for reference."""

from .null import NullCommand
from .ping import PingCommand
from .ship import ShipCommand
from .smecds import SmecdsCommand
from .ciaoruozi import CiaoruoziCommand
from .color import ColorCommand
from .sync import SyncCommand
from .diario import DiarioCommand
from .rage import RageCommand
from .dateparser import DateparserCommand
from .author import AuthorCommand
from .reminder import ReminderCommand
from .kvactive import KvactiveCommand
from .kv import KvCommand
from .kvroll import KvrollCommand
from .videoinfo import VideoinfoCommand
from .summon import SummonCommand
from .play import PlayCommand
from .skip import SkipCommand
from .playmode import PlaymodeCommand
from .videochannel import VideochannelCommand
from .missing import MissingCommand


__all__ = ["NullCommand", "PingCommand", "ShipCommand", "SmecdsCommand", "CiaoruoziCommand", "ColorCommand",
           "SyncCommand", "DiarioCommand", "RageCommand", "DateparserCommand", "AuthorCommand", "ReminderCommand",
           "KvactiveCommand", "KvCommand", "KvrollCommand", "VideoinfoCommand", "SummonCommand", "PlayCommand",
           "SkipCommand", "PlaymodeCommand", "VideochannelCommand", "MissingCommand"]



import discord
import typing
from ..utils import Command, Call


class VideochannelCommand(Command):

    command_name = "videochannel"
    command_description = "Converti il canale vocale in un canale video."
    command_syntax = "[channelname]"

    @classmethod
    async def discord(cls, call: Call):
        bot = call.interface_obj.client
        message: discord.Message = call.kwargs["message"]
        channel_name: str = call.args.optional(0)
        if channel_name:
            guild: typing.Optional[discord.Guild] = message.guild
            if guild is not None:
                channels: typing.List[discord.abc.GuildChannel] = guild.channels
            else:
                channels = bot.get_all_channels()
            matching_channels: typing.List[discord.VoiceChannel] = []
            for channel in channels:
                if isinstance(channel, discord.VoiceChannel):
                    if channel.name == channel_name:
                        matching_channels.append(channel)
            if len(matching_channels) == 0:
                await call.reply("⚠️ Non esiste alcun canale vocale con il nome specificato.")
                return
            elif len(matching_channels) > 1:
                await call.reply("⚠️ Esiste più di un canale vocale con il nome specificato.")
                return
            channel = matching_channels[0]
        else:
            author: discord.Member = message.author
            voice: typing.Optional[discord.VoiceState] = author.voice
            if voice is None:
                await call.reply("⚠️ Non sei connesso a nessun canale vocale!")
                return
            channel = voice.channel
        await call.reply(f"📹 Per entrare in modalità video, clicca qui: <https://discordapp.com/channels/{channel.guild.id}/{channel.id}>")



import typing
import discord
from ..network import Message, RequestSuccessful
from ..utils import Command, Call, NetworkHandler
from ..error import TooManyFoundError, NoneFoundError
if typing.TYPE_CHECKING:
    from ..bots import DiscordBot


class SkipMessage(Message):
    def __init__(self, guild_name: typing.Optional[str] = None):
        self.guild_name: typing.Optional[str] = guild_name


class SkipNH(NetworkHandler):
    message_type = SkipMessage

    @classmethod
    async def discord(cls, bot: "DiscordBot", message: SkipMessage):
        # Find the matching guild
        if message.guild_name:
            guild = bot.client.find_guild_by_name(message.guild_name)
        else:
            if len(bot.music_data) == 0:
                raise NoneFoundError("No voice clients active")
            if len(bot.music_data) > 1:
                raise TooManyFoundError("Multiple guilds found")
            guild = list(bot.music_data)[0]
        # Set the currently playing source as ended
        voice_client: discord.VoiceClient = bot.client.find_voice_client_by_guild(guild)
        if not voice_client.is_playing():
            raise NoneFoundError("Nothing to skip")
        # noinspection PyProtectedMember
        voice_client._player.stop()
        return RequestSuccessful()


class SkipCommand(Command):

    command_name = "skip"
    command_description = "Salta la canzone attualmente in riproduzione in chat vocale."
    command_syntax = "[ [guild] ]"

    network_handlers = [SkipNH]

    @classmethod
    async def common(cls, call: Call):
        guild, = call.args.match(r"(?:\[(.+)])?")
        await call.net_request(SkipMessage(guild), "discord")
        await call.reply(f"✅ Richiesto lo skip della canzone attuale.")



import typing
import discord
import asyncio
from ..utils import Command, Call, NetworkHandler
from ..network import Message, RequestSuccessful, RequestError
from ..error import NoneFoundError
if typing.TYPE_CHECKING:
    from ..bots import DiscordBot


loop = asyncio.get_event_loop()


class SummonMessage(Message):
    def __init__(self, channel_identifier: typing.Union[int, str],
                 guild_identifier: typing.Optional[typing.Union[int, str]] = None):
        self.channel_name = channel_identifier
        self.guild_identifier = guild_identifier


class SummonNH(NetworkHandler):
    message_type = SummonMessage

    @classmethod
    async def discord(cls, bot: "DiscordBot", message: SummonMessage):
        """Handle a summon Royalnet request. That is, join a voice channel, or move to a different one if that is not possible."""
        channel = bot.client.find_channel_by_name(message.channel_name)
        if not isinstance(channel, discord.VoiceChannel):
            raise NoneFoundError("Channel is not a voice channel")
        loop.create_task(bot.client.vc_connect_or_move(channel))
        return RequestSuccessful()


class SummonCommand(Command):

    command_name = "summon"
    command_description = "Evoca il bot in un canale vocale."
    command_syntax = "[channelname]"

    network_handlers = [SummonNH]

    @classmethod
    async def common(cls, call: Call):
        channel_name: str = call.args[0].lstrip("#")
        response: typing.Union[RequestSuccessful, RequestError] = await call.net_request(SummonMessage(channel_name), "discord")
        response.raise_on_error()
        await call.reply(f"✅ Mi sono connesso in [c]#{channel_name}[/c].")

    @classmethod
    async def discord(cls, call: Call):
        bot = call.interface_obj.client
        message: discord.Message = call.kwargs["message"]
        channel_name: str = call.args.optional(0)
        if channel_name:
            guild: typing.Optional[discord.Guild] = message.guild
            if guild is not None:
                channels: typing.List[discord.abc.GuildChannel] = guild.channels
            else:
                channels = bot.get_all_channels()
            matching_channels: typing.List[discord.VoiceChannel] = []
            for channel in channels:
                if isinstance(channel, discord.VoiceChannel):
                    if channel.name == channel_name:
                        matching_channels.append(channel)
            if len(matching_channels) == 0:
                await call.reply("⚠️ Non esiste alcun canale vocale con il nome specificato.")
                return
            elif len(matching_channels) > 1:
                await call.reply("⚠️ Esiste più di un canale vocale con il nome specificato.")
                return
            channel = matching_channels[0]
        else:
            author: discord.Member = message.author
            voice: typing.Optional[discord.VoiceState] = author.voice
            if voice is None:
                await call.reply("⚠️ Non sei connesso a nessun canale vocale!")
                return
            channel = voice.channel
        await bot.vc_connect_or_move(channel)
        await call.reply(f"✅ Mi sono connesso in [c]#{channel.name}[/c].")



from .royals import Royal
from .telegram import Telegram
from .diario import Diario
from .aliases import Alias
from .activekvgroup import ActiveKvGroup
from .keyvalue import Keyvalue
from .keygroup import Keygroup
from .discord import Discord

__all__ = ["Royal", "Telegram", "Diario", "Alias", "ActiveKvGroup", "Keyvalue", "Keygroup", "Discord"]



import datetime
import dateparser
from ..utils import Command, Call


class DateparserCommand(Command):

    command_name = "dateparser"
    command_description = "Legge e comprende la data inserita."
    command_syntax = "(data)"

    @classmethod
    async def common(cls, call: Call):
        text = call.args.joined(require_at_least=1)
        date: datetime.datetime = dateparser.parse(text)
        if date is None:
            await call.reply("🕕 La data inserita non è valida.")
            return
        await call.reply(f"🕐 La data inserita è {date.strftime('%Y-%m-%d %H:%M:%S')}")



import typing
import asyncio
from ..utils import Command, Call, NetworkHandler
from ..network import Message, RequestSuccessful
from ..error import NoneFoundError, TooManyFoundError
from ..audio import Playlist, Pool
if typing.TYPE_CHECKING:
    from ..bots import DiscordBot


loop = asyncio.get_event_loop()


class PlaymodeMessage(Message):
    def __init__(self, mode_name: str, guild_name: typing.Optional[str] = None):
        self.mode_name: str = mode_name
        self.guild_name: typing.Optional[str] = guild_name


class PlaymodeNH(NetworkHandler):
    message_type = PlaymodeMessage

    @classmethod
    async def discord(cls, bot: "DiscordBot", message: PlaymodeMessage):
        """Handle a playmode Royalnet request. That is, change current PlayMode."""
        # Find the matching guild
        if message.guild_name:
            guild = bot.client.find_guild(message.guild_name)
        else:
            if len(bot.music_data) == 0:
                raise NoneFoundError("No voice clients active")
            if len(bot.music_data) > 1:
                raise TooManyFoundError("Multiple guilds found")
            guild = list(bot.music_data)[0]
        # Delete the previous PlayMode, if it exists
        if bot.music_data[guild] is not None:
            bot.music_data[guild].delete()
        # Create the new PlayMode
        if message.mode_name == "playlist":
            bot.music_data[guild] = Playlist()
        elif message.mode_name == "pool":
            bot.music_data[guild] = Pool()
        else:
            raise ValueError("No such PlayMode")
        return RequestSuccessful()


class PlaymodeCommand(Command):
    command_name = "playmode"
    command_description = "Cambia modalità di riproduzione per la chat vocale."
    command_syntax = "[ [guild] ] (mode)"

    network_handlers = [PlaymodeNH]

    @classmethod
    async def common(cls, call: Call):
        guild, mode_name = call.args.match(r"(?:\[(.+)])?\s*(\S+)\s*")
        await call.net_request(PlaymodeMessage(mode_name, guild), "discord")
        await call.reply(f"✅ Richiesto di passare alla modalità di riproduzione [c]{mode_name}[/c].")



from ..utils import Command, Call, asyncify
from ..database.tables import Royal, Alias


class DebugCreateCommand(Command):

    command_name = "debug_create"
    command_description = "Crea un nuovo account Royalnet"
    command_syntax = "(newusername)"

    require_alchemy_tables = {Royal, Alias}

    @classmethod
    async def common(cls, call: Call):
        royal = call.alchemy.Royal(username=call.args[0], role="Member")
        call.session.add(royal)
        alias = call.alchemy.Alias(royal=royal, alias=royal.username.lower())
        call.session.add(alias)
        await asyncify(call.session.commit)
        await call.reply(f"✅ Utente {royal} creato!")



import typing
import pickle
from ..error import RoyalnetError


class Message:
    """A message sent through the Royalnet."""
    def __repr__(self):
        return f"<{self.__class__.__name__}>"


class IdentifySuccessfulMessage(Message):
    """The Royalnet identification step was successful."""


class ServerErrorMessage(Message):
    """Something went wrong in the connection to the :py:class:`royalnet.network.RoyalnetServer`."""
    def __init__(self, reason):
        super().__init__()
        self.reason = reason


class InvalidSecretEM(ServerErrorMessage):
    """The sent secret was incorrect.

     This message terminates connection to the :py:class:`royalnet.network.RoyalnetServer`."""


class InvalidPackageEM(ServerErrorMessage):
    """The sent :py:class:`royalnet.network.Package` was invalid."""


class InvalidDestinationEM(InvalidPackageEM):
    """The :py:class:`royalnet.network.Package` destination was invalid or not found."""


class Reply(Message):
    """A reply to a request sent through the Royalnet."""

    def raise_on_error(self) -> None:
        """If the reply is an error, raise an error, otherwise, do nothing.

        Raises:
            A :py:exc:`RoyalnetError`, if the Reply is an error, otherwise, nothing."""
        raise NotImplementedError()


class RequestSuccessful(Reply):
    """The sent request was successful."""

    def raise_on_error(self) -> None:
        """If the reply is an error, raise an error, otherwise, do nothing.

        Does nothing."""
        pass


class RequestError(Reply):
    """The sent request wasn't successful."""

    def __init__(self, exc: typing.Optional[Exception] = None):
        """Create a RequestError.

        Parameters:
             exc: The exception that caused the error in the request."""
        try:
            pickle.dumps(exc)
        except TypeError:
            self.exc: Exception = Exception(repr(exc))
        else:
            self.exc = exc

    def raise_on_error(self) -> None:
        """If the reply is an error, raise an error, otherwise, do nothing.

        Raises:
            Always raises a :py:exc:`royalnet.error.RoyalnetError`, containing the exception that caused the error."""
        raise RoyalnetError(exc=self.exc)



from ..utils import Command, Call


class NullCommand(Command):

    command_name = "null"
    command_description = "Non fa nulla."
    command_syntax = ""

    @classmethod
    async def common(cls, call: Call):
        pass



import random
from ..database.tables import ActiveKvGroup, Royal, Keygroup, Keyvalue
from ..utils import Command, Call, asyncify, plusformat


class KvrollCommand(Command):

    command_name = "kvroll"
    command_description = "Lancia 1d20, poi aggiungici il valore della kv selezionata."
    command_syntax = "(chiave) [modifier]"

    require_alchemy_tables = {ActiveKvGroup, Royal, Keyvalue, Keygroup}

    @classmethod
    async def common(cls, call: Call):
        key = call.args[0].lower()
        normal_mod_str = call.args.optional(1, 0)
        try:
            normal_modifier = int(normal_mod_str)
        except ValueError:
            await call.reply("⚠️ Il modificatore specificato non è un numero.")
            return
        author = await call.get_author(error_if_none=True)
        active = await asyncify(call.session.query(call.alchemy.ActiveKvGroup).filter_by(royal=author).one_or_none)
        if active is None:
            await call.reply("⚠️ Devi prima attivare un gruppo con il comando [c]kvactive[/c]!")
            return
        keyvalue = await asyncify(call.session.query(call.alchemy.Keyvalue).filter_by(group=active.group, key=key).one_or_none)
        if keyvalue is None:
            await call.reply("⚠️ La chiave specificata non esiste.")
            return
        try:
            kv_modifier = int(keyvalue.value)
        except ValueError:
            await call.reply("⚠️ Il valore della chiave specificata non è un numero.")
            return
        roll = random.randrange(1, 21)
        result = roll + kv_modifier + normal_modifier
        await call.reply(f"🎲 {roll}{plusformat(kv_modifier)}{plusformat(normal_modifier)} = [b]{result}[/b]")



from ..database.tables import ActiveKvGroup, Royal, Keygroup
from ..utils import Command, Call, asyncify


class KvactiveCommand(Command):

    command_name = "kvactive"
    command_description = "Seleziona un gruppo di valori kv."
    command_syntax = "(nomegruppo)"

    require_alchemy_tables = {ActiveKvGroup, Royal, Keygroup}

    @classmethod
    async def common(cls, call: Call):
        group_name = call.args[0].lower()
        author = await call.get_author(error_if_none=True)
        active = await asyncify(call.session.query(call.alchemy.ActiveKvGroup).filter_by(royal=author).one_or_none)
        if active is None:
            group = await asyncify(call.session.query(call.alchemy.Keygroup).filter_by(group_name=group_name).one_or_none)
            if group is None:
                group = call.alchemy.Keygroup(group_name=group_name)
                call.session.add(group)
            active = call.alchemy.ActiveKvGroup(royal=author, group=group)
            call.session.add(active)
        else:
            group = await asyncify(call.session.query(call.alchemy.Keygroup).filter_by(group_name=group_name).one_or_none)
            if group is None:
                group = call.alchemy.Keygroup(group_name=group_name)
                call.session.add(group)
            active.group = group
        await asyncify(call.session.commit)
        await call.reply(f"✅ Hai attivato il gruppo [b]{group_name}[/b].")



import telegram
from telegram.utils.request import Request
import asyncio
import typing
import logging as _logging
from .generic import GenericBot
from ..commands import NullCommand
from ..utils import asyncify, Call, Command
from ..error import UnregisteredError, InvalidConfigError
from ..network import Message, RoyalnetConfig, Reply
from ..database import DatabaseConfig

loop = asyncio.get_event_loop()
log = _logging.getLogger(__name__)


class TelegramConfig:
    """The specific configuration to be used for :py:class:`royalnet.database.TelegramBot`."""
    def __init__(self, token: str):
        self.token: str = token


class TelegramBot(GenericBot):
    """A bot that connects to `Telegram <https://telegram.org/>`_."""
    interface_name = "telegram"

    def _init_client(self):
        """Create the :py:class:`telegram.Bot`, and set the starting offset."""
        # https://github.com/python-telegram-bot/python-telegram-bot/issues/341
        request = Request(5)
        self.client = telegram.Bot(self._telegram_config.token, request=request)
        self._offset: int = -100

    def _call_factory(self) -> typing.Type[Call]:
        # noinspection PyMethodParameters
        class TelegramCall(Call):
            interface_name = self.interface_name
            interface_obj = self
            interface_prefix = "/"

            alchemy = self.alchemy

            async def reply(call, text: str):
                escaped_text = text.replace("<", "&lt;") \
                                   .replace(">", "&gt;") \
                                   .replace("[b]", "<b>") \
                                   .replace("[/b]", "</b>") \
                                   .replace("[i]", "<i>") \
                                   .replace("[/i]", "</i>") \
                                   .replace("[u]", "<b>") \
                                   .replace("[/u]", "</b>") \
                                   .replace("[c]", "<code>") \
                                   .replace("[/c]", "</code>") \
                                   .replace("[p]", "<pre>") \
                                   .replace("[/p]", "</pre>")
                await asyncify(call.channel.send_message, escaped_text, parse_mode="HTML")

            async def net_request(call, message: Message, destination: str):
                if self.network is None:
                    raise InvalidConfigError("Royalnet is not enabled on this bot")
                response: Reply = await self.network.request(message, destination)
                response.raise_on_error()
                return response

            async def get_author(call, error_if_none=False):
                update: telegram.Update = call.kwargs["update"]
                user: telegram.User = update.effective_user
                if user is None:
                    if error_if_none:
                        raise UnregisteredError("No author for this message")
                    return None
                query = call.session.query(self.master_table)
                for link in self.identity_chain:
                    query = query.join(link.mapper.class_)
                query = query.filter(self.identity_column == user.id)
                result = await asyncify(query.one_or_none)
                if result is None and error_if_none:
                    raise UnregisteredError("Author is not registered")
                return result

        return TelegramCall

    def __init__(self, *,
                 telegram_config: TelegramConfig,
                 royalnet_config: typing.Optional[RoyalnetConfig] = None,
                 database_config: typing.Optional[DatabaseConfig] = None,
                 command_prefix: str = "/",
                 commands: typing.List[typing.Type[Command]] = None,
                 missing_command: typing.Type[Command] = NullCommand,
                 error_command: typing.Type[Command] = NullCommand):
        super().__init__(royalnet_config=royalnet_config,
                         database_config=database_config,
                         command_prefix=command_prefix,
                         commands=commands,
                         missing_command=missing_command,
                         error_command=error_command)
        self._telegram_config = telegram_config
        self._init_client()

    async def _handle_update(self, update: telegram.Update):
        # Skip non-message updates
        if update.message is None:
            return
        message: telegram.Message = update.message
        text: str = message.text
        # Try getting the caption instead
        if text is None:
            text: str = message.caption
        # No text or caption, ignore the message
        if text is None:
            return
        # Find and clean parameters
        command_text, *parameters = text.split(" ")
        command_name = command_text.replace(f"@{self.client.username}", "").lower()
        # Call the command
        await self.call(command_name, update.message.chat, parameters, update=update)

    async def run(self):
        while True:
            # Get the latest 100 updates
            try:
                last_updates: typing.List[telegram.Update] = await asyncify(self.client.get_updates, offset=self._offset, timeout=60)
            except telegram.error.TimedOut:
                continue
            # Handle updates
            for update in last_updates:
                # noinspection PyAsyncCall
                loop.create_task(self._handle_update(update))
            # Recalculate offset
            try:
                self._offset = last_updates[-1].update_id + 1
            except IndexError:
                pass

    @property
    def botfather_command_string(self) -> str:
        string = ""
        for command_key in self.commands:
            command = self.commands[command_key]
            string += f"{command.command_name} - {command.command_description}\n"
        return string



import re
from sqlalchemy import Column, \
                       Integer, \
                       Text, \
                       Boolean, \
                       DateTime, \
                       ForeignKey, \
                       String
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declared_attr
# noinspection PyUnresolvedReferences
from .royals import Royal


class Diario:
    __tablename__ = "diario"

    @declared_attr
    def diario_id(self):
        return Column(Integer, primary_key=True)

    @declared_attr
    def creator_id(self):
        return Column(Integer, ForeignKey("royals.uid"), nullable=False)

    @declared_attr
    def quoted_account_id(self):
        return Column(Integer, ForeignKey("royals.uid"))

    @declared_attr
    def quoted(self):
        return Column(String)

    @declared_attr
    def text(self):
        return Column(Text)

    @declared_attr
    def context(self):
        return Column(Text)

    @declared_attr
    def timestamp(self):
        return Column(DateTime, nullable=False)

    @declared_attr
    def media_url(self):
        return Column(String)

    @declared_attr
    def spoiler(self):
        return Column(Boolean, default=False)

    @declared_attr
    def creator(self):
        return relationship("Royal", foreign_keys=self.creator_id, backref="diario_created")

    @declared_attr
    def quoted_account(self):
        return relationship("Royal", foreign_keys=self.quoted_account_id, backref="diario_quoted")

    def __repr__(self):
        return f"<Diario diario_id={self.diario_id} creator_id={self.creator_id} quoted_account_id={self.quoted_account_id} quoted={self.quoted} text={self.text} context={self.context} timestamp={self.timestamp} media_url={self.media_url} spoiler={self.spoiler}>"

    def __str__(self):
        # TODO: support media_url
        text = f"Riga #{self.diario_id}"
        text += f" (salvata da {str(self.creator)}"
        text += f" il {self.timestamp.strftime('%Y-%m-%d %H:%M')}):\n"
        if self.media_url is not None:
            text += f"{self.media_url}\n"
        if self.text is not None:
            if self.spoiler:
                hidden = re.sub(r"\w", "█", self.text)
                text += f"\"{hidden}\"\n"
            else:
                text += f"[b]\"{self.text}\"[/b]\n"
        if self.quoted_account is not None:
            text += f" —{str(self.quoted_account)}"
        elif self.quoted is not None:
            text += f" —{self.quoted}"
        else:
            text += f" —Anonimo"
        if self.context:
            text += f", [i]{self.context}[/i]"
        return text



import royalnet
import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="royalnet",
    version=royalnet.version,
    author="Stefano Pigozzi",
    author_email="ste.pigozzi@gmail.com",
    description="The great bot network of the Royal Games community",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Steffo99/royalnet",
    packages=setuptools.find_packages(),
    install_requires=["python-telegram-bot>=11.1.0",
                      "websockets>=7.0",
                      "psycopg2-binary>=2.8",
                      "aiohttp>=3.5.4",
                      "sqlalchemy>=1.3.2",
                      "Markdown>=3.1",
                      "dateparser>=0.7.1",
                      "discord.py>=1.0.1",
                      "youtube_dl>=2019.4.24",
                      "ffmpeg-python>=0.1.17",
                      "PyNaCl>=1.3.0"],
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Topic :: Internet",
        "Topic :: Database",
        "Topic :: Multimedia :: Sound/Audio",
        "Topic :: Multimedia :: Video",
        "License :: OSI Approved :: MIT License"
    ]
)



"""Miscellaneous useful functions and classes."""

from .asyncify import asyncify
from .call import Call
from .command import Command
from .commandargs import CommandArgs
from .safeformat import safeformat
from .classdictjanitor import cdj
from .sleepuntil import sleep_until
from .plusformat import plusformat
from .networkhandler import NetworkHandler
from .safefilename import safefilename

__all__ = ["asyncify", "Call", "Command", "safeformat", "cdj", "sleep_until", "plusformat", "CommandArgs",
           "NetworkHandler", "safefilename"]



class NoneFoundError(Exception):
    """The element that was being looked for was not found."""


class TooManyFoundError(Exception):
    """Multiple elements matching the request were found, and only one was expected."""


class UnregisteredError(Exception):
    """The command required a registered user, and the user was not registered."""


class UnsupportedError(Exception):
    """The command is not supported for the specified interface."""


class InvalidInputError(Exception):
    """The command has received invalid input and cannot complete."""


class InvalidConfigError(Exception):
    """The bot has not been configured correctly, therefore the command can not function."""


class RoyalnetError(Exception):
    """An error was raised while handling the Royalnet request.
    This exception contains the exception that was raised during the handling."""
    def __init__(self, exc: Exception):
        self.exc: Exception = exc


class ExternalError(Exception):
    """Something went wrong in a non-Royalnet component and the command execution cannot be completed."""



import typing
import logging as _logging
import os
from youtube_dl import YoutubeDL

log = _logging.getLogger(__name__)


class DownloaderError(Exception):
    pass


class InterruptDownload(DownloaderError):
    """Raised from a progress_hook to interrupt the video download."""


class YtdlFile:
    """A wrapper around a youtube_dl downloaded file."""

    ytdl_args = {
        "logger": log,  # Log messages to a logging.Logger instance.
        "quiet": True,  # Do not print messages to stdout.
        "noplaylist": True,  # Download single video instead of a playlist if in doubt.
        "no_warnings": True,  # Do not print out anything for warnings.
    }

    def __init__(self, info: "YtdlInfo", outtmpl="%(title)s-%(id)s.%(ext)s", **ytdl_args):
        self.info: "YtdlInfo" = info
        self.video_filename: str
        # Create a local args copy
        ytdl_args["outtmpl"] = outtmpl
        self.ytdl_args = {**self.ytdl_args, **ytdl_args}
        # Create the ytdl
        ytdl = YoutubeDL(ytdl_args)
        # Find the file name
        self.video_filename = ytdl.prepare_filename(self.info.__dict__)
        # Download the file
        ytdl.download([self.info.webpage_url])
        # Final checks
        assert os.path.exists(self.video_filename)

    def __repr__(self):
        return f"<YtdlFile {self.video_filename}>"

    @staticmethod
    def create_from_url(url, outtmpl="%(title)s-%(id)s.%(ext)s", **ytdl_args) -> typing.List["YtdlFile"]:
        info_list = YtdlInfo.create_from_url(url)
        return [info.download(outtmpl, **ytdl_args) for info in info_list]

    def _stop_download(self):
        raise InterruptDownload()

    def delete_video_file(self):
        # TODO: _might_ be unsafe, test this
        os.remove(self.video_filename)


class YtdlInfo:
    """A wrapper around youtube_dl extracted info."""

    def __init__(self, info: typing.Dict[str, typing.Any]):
        """Create a YtdlInfo from the dict returned by the :py:func:`youtube_dl.YoutubeDL.extract_info` function.

        Warning:
            Does not download the info, for that use :py:func:`royalnet.audio.YtdlInfo.create_from_url`."""
        self.id: typing.Optional[str] = info.get("id")
        self.uploader: typing.Optional[str] = info.get("uploader")
        self.uploader_id: typing.Optional[str] = info.get("uploader_id")
        self.uploader_url: typing.Optional[str] = info.get("uploader_url")
        self.channel_id: typing.Optional[str] = info.get("channel_id")
        self.channel_url: typing.Optional[str] = info.get("channel_url")
        self.upload_date: typing.Optional[str] = info.get("upload_date")
        self.license: typing.Optional[str] = info.get("license")
        self.creator: typing.Optional[...] = info.get("creator")
        self.title: typing.Optional[str] = info.get("title")
        self.alt_title: typing.Optional[...] = info.get("alt_title")
        self.thumbnail: typing.Optional[str] = info.get("thumbnail")
        self.description: typing.Optional[str] = info.get("description")
        self.categories: typing.Optional[typing.List[str]] = info.get("categories")
        self.tags: typing.Optional[typing.List[str]] = info.get("tags")
        self.subtitles: typing.Optional[typing.Dict[str, typing.List[typing.Dict[str, str]]]] = info.get("subtitles")
        self.automatic_captions: typing.Optional[dict] = info.get("automatic_captions")
        self.duration: typing.Optional[int] = info.get("duration")
        self.age_limit: typing.Optional[int] = info.get("age_limit")
        self.annotations: typing.Optional[...] = info.get("annotations")
        self.chapters: typing.Optional[...] = info.get("chapters")
        self.webpage_url: typing.Optional[str] = info.get("webpage_url")
        self.view_count: typing.Optional[int] = info.get("view_count")
        self.like_count: typing.Optional[int] = info.get("like_count")
        self.dislike_count: typing.Optional[int] = info.get("dislike_count")
        self.average_rating: typing.Optional[...] = info.get("average_rating")
        self.formats: typing.Optional[list] = info.get("formats")
        self.is_live: typing.Optional[bool] = info.get("is_live")
        self.start_time: typing.Optional[float] = info.get("start_time")
        self.end_time: typing.Optional[float] = info.get("end_time")
        self.series: typing.Optional[str] = info.get("series")
        self.season_number: typing.Optional[int] = info.get("season_number")
        self.episode_number: typing.Optional[int] = info.get("episode_number")
        self.track: typing.Optional[...] = info.get("track")
        self.artist: typing.Optional[...] = info.get("artist")
        self.extractor: typing.Optional[str] = info.get("extractor")
        self.webpage_url_basename: typing.Optional[str] = info.get("webpage_url_basename")
        self.extractor_key: typing.Optional[str] = info.get("extractor_key")
        self.playlist: typing.Optional[str] = info.get("playlist")
        self.playlist_index: typing.Optional[int] = info.get("playlist_index")
        self.thumbnails: typing.Optional[typing.List[typing.Dict[str, str]]] = info.get("thumbnails")
        self.display_id: typing.Optional[str] = info.get("display_id")
        self.requested_subtitles: typing.Optional[...] = info.get("requested_subtitles")
        self.requested_formats: typing.Optional[tuple] = info.get("requested_formats")
        self.format: typing.Optional[str] = info.get("format")
        self.format_id: typing.Optional[str] = info.get("format_id")
        self.width: typing.Optional[int] = info.get("width")
        self.height: typing.Optional[int] = info.get("height")
        self.resolution: typing.Optional[...] = info.get("resolution")
        self.fps: typing.Optional[int] = info.get("fps")
        self.vcodec: typing.Optional[str] = info.get("vcodec")
        self.vbr: typing.Optional[int] = info.get("vbr")
        self.stretched_ratio: typing.Optional[...] = info.get("stretched_ratio")
        self.acodec: typing.Optional[str] = info.get("acodec")
        self.abr: typing.Optional[int] = info.get("abr")
        self.ext: typing.Optional[str] = info.get("ext")

    @staticmethod
    def create_from_url(url, **ytdl_args) -> typing.List["YtdlInfo"]:
        # So many redundant options!
        ytdl = YoutubeDL({
            "logger": log,  # Log messages to a logging.Logger instance.
            "quiet": True,  # Do not print messages to stdout.
            "noplaylist": True,  # Download single video instead of a playlist if in doubt.
            "no_warnings": True,  # Do not print out anything for warnings.
            **ytdl_args
        })
        first_info = ytdl.extract_info(url=url, download=False)
        # If it is a playlist, create multiple videos!
        if "entries" in first_info:
            return [YtdlInfo(second_info) for second_info in first_info["entries"]]
        return [YtdlInfo(first_info)]

    def download(self, outtmpl="%(title)s-%(id)s.%(ext)s", **ytdl_args) -> YtdlFile:
        return YtdlFile(self, outtmpl, **ytdl_args)

    def __repr__(self):
        if self.title:
            return f"<YtdlInfo of {self.title}>"
        if self.webpage_url:
            return f"<YtdlInfo for {self.webpage_url}>"
        return f"<YtdlInfo id={self.id} ...>"

    def __str__(self):
        if self.title:
            return self.title
        if self.webpage_url:
            return self.webpage_url
        return self.id



import typing
from telegram import Update, User
from discord import Message, Member
from ..utils import Command, Call, asyncify
from ..error import UnsupportedError
from ..database.tables import Royal, Telegram, Discord


class SyncCommand(Command):

    command_name = "sync"
    command_description = "Connetti il tuo account attuale a Royalnet!"
    command_syntax = "(royalnetusername)"

    require_alchemy_tables = {Royal, Telegram, Discord}

    @classmethod
    async def common(cls, call: Call):
        raise UnsupportedError()

    @classmethod
    async def telegram(cls, call: Call):
        update: Update = call.kwargs["update"]
        # Find the user
        user: typing.Optional[User] = update.effective_user
        if user is None:
            raise ValueError("Trying to sync a None user.")
        # Find the Royal
        royal = await asyncify(call.session.query(call.alchemy.Royal).filter_by(username=call.args[0]).one_or_none)
        if royal is None:
            await call.reply("⚠️ Non esiste alcun account Royalnet con quel nome. Ricorda, gli username sono [b]case-sensitive[/b]!")
            return
        # Find if the user is already synced
        telegram = await asyncify(call.session.query(call.alchemy.Telegram).filter_by(tg_id=user.id).one_or_none)
        if telegram is None:
            # Create a Telegram to connect to the Royal
            # Avatar is WIP
            telegram = call.alchemy.Telegram(royal=royal,
                                             tg_id=user.id,
                                             first_name=user.first_name,
                                             last_name=user.last_name,
                                             username=user.username)
            call.session.add(telegram)
            await call.reply(f"✅ Connessione completata: {str(royal)} ⬌ {str(telegram)}")
        else:
            # Update the Telegram data
            telegram.first_name = user.first_name
            telegram.last_name = user.last_name
            telegram.username = user.username
            await call.reply(f"✅ Connessione aggiornata: {str(royal)} ⬌ {str(telegram)}")
        # Commit the session
        await asyncify(call.session.commit)

    @classmethod
    async def discord(cls, call: Call):
        message: Message = call.kwargs["message"]
        user: typing.Optional[Member] = message.author
        # Find the Royal
        royal = await asyncify(call.session.query(call.alchemy.Royal).filter_by(username=call.args[0]).one_or_none)
        if royal is None:
            await call.reply("⚠️ Non esiste alcun account Royalnet con quel nome. Ricorda, gli username sono [b]case-sensitive[/b]!")
            return
        # Find if the user is already synced
        discord = await asyncify(call.session.query(call.alchemy.Discord).filter_by(discord_id=user.id).one_or_none)
        if discord is None:
            # Create a Discord to connect to the Royal
            discord = call.alchemy.Discord(royal=royal,
                                           discord_id=user.id,
                                           username=user.name,
                                           discriminator=user.discriminator,
                                           avatar_hash=user.avatar)
            call.session.add(discord)
            await call.reply(f"✅ Connessione completata: {str(royal)} ⬌ {str(discord)}")
        else:
            # Update the Discord data
            discord.username = user.name
            discord.discriminator = user.discriminator
            discord.avatar_hash = user.avatar
            await call.reply(f"✅ Connessione aggiornata: {str(royal)} ⬌ {str(discord)}")
        # Commit the session
        await asyncify(call.session.commit)



"""Royalnet realated classes."""

from .messages import Message, ServerErrorMessage, InvalidSecretEM, InvalidDestinationEM, InvalidPackageEM, RequestSuccessful, RequestError, Reply
from .packages import Package
from .royalnetlink import RoyalnetLink, NetworkError, NotConnectedError, NotIdentifiedError
from .royalnetserver import RoyalnetServer
from .royalnetconfig import RoyalnetConfig

__all__ = ["Message",
           "ServerErrorMessage",
           "InvalidSecretEM",
           "InvalidDestinationEM",
           "InvalidPackageEM",
           "RoyalnetLink",
           "NetworkError",
           "NotConnectedError",
           "NotIdentifiedError",
           "Package",
           "RoyalnetServer",
           "RequestSuccessful",
           "RequestError",
           "RoyalnetConfig",
           "Reply"]



def plusformat(i: int) -> str:
    """Convert an :py:class:`int` to a string, adding a plus if they are greater than 0.

    Parameters:
        i: the :py:class:`int` to convert.

    Returns:
        The resulting :py:class:`str`."""
    if i >= 0:
        return f"+{i}"
    return str(i)
